__project__ = 'hermes-audio-server'
__description__ = 'An open source implementation of the audio server part of the Hermes protocol'
__keywords__ = 'hermes snips python3 rhasspy audio-player audio-recorder audio-server snips-audio-server hermes-protocol voice voice-control'
__author__ = 'Koen Vervloesem'
__email__ = 'koen@vervloesem.eu'
__license__ = 'MIT License'
__github_url__ = 'https://github.com/koenvervloesem/hermes-audio-server'
__doc_url__ = 'https://github.com/koenvervloesem/hermes-audio-server/README.md'
__tracker_url__ = 'https://github.com/koenvervloesem/hermes-audio-server/issues'
__player__ = 'hermes-audio-player'
__recorder__ = 'hermes-audio-recorder'
__version__ = '0.0.2'



"""Module with the Hermes audio player class."""
import io
import json
import wave

from humanfriendly import format_size
from paho.mqtt.client import Client
import pyaudio

from hermes_audio_server.mqtt import connect

PLAY_BYTES = 'hermes/audioServer/{}/playBytes/+'
PLAY_FINISHED = 'hermes/audioServer/{}/playFinished'
CHUNK = 256


class AudioPlayer:
    """This class creates an MQTT client that acts as an audio player for the
    Hermes protocol.
    """

    def __init__(self, config, verbose):
        """Initialize a Hermes audio player.

        Args:
            config (:class:`.ServerConfig`): The configuration of
                the Hermes audio player.
        """
        self.config = config
        self.verbose = verbose
        self.mqtt = Client()
        print('Audio player created on site {}.'.format(self.config.site))

        self.audio = pyaudio.PyAudio()
        self.audio_out = self.audio.get_default_output_device_info()['name']
        print('Connected to audio output {}.'.format(self.audio_out))

        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect

        connect(self.mqtt, self.config.mqtt)

    def start(self):
        """Start the event loop to the MQTT broker so the audio player starts
        listening to MQTT topics and the callback methods are called.
        """
        self.mqtt.loop_forever()

    def on_connect(self, client, userdata, flags, result_code):
        print('Connected to MQTT broker {}:{}'
              ' with result code {}.'.format(self.config.mqtt.host,
                                             self.config.mqtt.port,
                                             result_code))

        play_bytes = PLAY_BYTES.format(self.config.site)
        self.mqtt.subscribe(play_bytes)
        self.mqtt.message_callback_add(play_bytes, self.on_play_bytes)
        print('Subscribed to {}.'.format(play_bytes))

    def on_disconnect(self, client, userdata, flags, result_code):
        print('Disconnected with result code {}.'.format(result_code))

    def on_play_bytes(self, client, userdata, message):
        request_id = message.topic.split('/')[4]
        length = format_size(len(message.payload), binary=True)
        print('Received an audio message of length {}'
              ' with request id {} on site {}.'.format(length,
                                                       request_id,
                                                       self.config.site))

        with io.BytesIO(message.payload) as wav_buffer:
            with wave.open(wav_buffer, 'rb') as wav:
                sample_width = wav.getsampwidth()
                sample_format = self.audio.get_format_from_width(sample_width)
                n_channels = wav.getnchannels()
                frame_rate = wav.getframerate()

                if self.verbose:
                    print('Sample width: {}'.format(sample_width))
                    print('Channels: {}'.format(n_channels))
                    print('Frame rate: {}'.format(frame_rate))

                stream = self.audio.open(format=sample_format,
                                         channels=n_channels,
                                         rate=frame_rate,
                                         output=True)

                data = wav.readframes(CHUNK)

                while len(data) > 0:
                    stream.write(data)
                    data = wav.readframes(CHUNK)

                stream.stop_stream()
                stream.close()

        self.mqtt.publish(PLAY_FINISHED.format(self.config.site),
                          json.dumps({'id': request_id,
                                      'siteId': self.config.site}))
        print('Finished playing audio message with id {}'
              ' on device {} on site {}'.format(request_id,
                                                self.audio_out,
                                                self.config.site))



"""Helper functions to use the Paho MQTT library with the MQTT broker defined
in a :class:`.MQTTConfig` object.
"""


def connect(client, mqtt_config, keepalive=60, bind_address=''):
    """Connect to an MQTT broker with the MQTT connection settings defined in
    an :class:`.MQTTConfig` object.

    Args:
        client (:class:`paho.mqtt.client.Client`): The MQTT client object.
        mqtt_config (:class:`.MQTTConfig`): The MQTT connection settings.
        keepalive (int, optional): The maximum period in seconds allowed
            between communications with the broker. Defaults to 60.
        bind_address (str, optional): The IP address of a local network
            interface to bind this client to, assuming multiple interfaces
            exist. Defaults to ''.
    """
    # Set up MQTT authentication.
    if mqtt_config.auth.enabled:
        client.username_pw_set(mqtt_config.auth.username,
                               mqtt_config.auth.password)

    # Set up an MQTT TLS connection.
    if mqtt_config.tls.enabled:
        client.tls_set(ca_certs=mqtt_config.tls.ca_certs,
                       certfile=mqtt_config.tls.client_cert,
                       keyfile=mqtt_config.tls.client_key)

    client.connect(mqtt_config.host, mqtt_config.port, keepalive, bind_address)



"""Module with the Hermes audio recorder class."""
import io
from threading import Thread
import wave

from paho.mqtt.client import Client
import pyaudio

from hermes_audio_server.mqtt import connect

AUDIO_FRAME = 'hermes/audioServer/{}/audioFrame'
CHANNELS = 1
CHUNK = 256
FRAME_RATE = 16000
SAMPLE_WIDTH = 2


# TODO: Call stream.stop_stream() and stream.close()
class AudioRecorder:
    """This class creates an MQTT client that acts as an audio recorder for the
    Hermes protocol.
    """

    def __init__(self, config, verbose):
        """Initialize a Hermes audio recorder.

        Args:
            config (:class:`.ServerConfig`): The configuration of
                the Hermes audio recorder.
        """
        self.config = config
        self.verbose = verbose
        self.mqtt = Client()
        print('Audio recorder created on site {}.'.format(self.config.site))

        self.audio = pyaudio.PyAudio()
        self.audio_in = self.audio.get_default_input_device_info()['name']
        print('Connected to audio input {}.'.format(self.audio_in))

        self.mqtt.on_connect = self.on_connect
        self.mqtt.on_disconnect = self.on_disconnect

        connect(self.mqtt, self.config.mqtt)

    def start(self):
        """Start the event loop to the MQTT broker and start the audio
        recording."""
        Thread(target=self.send_audio_frames, daemon=True).start()
        self.mqtt.loop_forever()

    def on_connect(self, client, userdata, flags, result_code):
        print('Connected to MQTT broker {}:{}'
              ' with result code {}.'.format(self.config.mqtt.host,
                                             self.config.mqtt.port,
                                             result_code))

    def on_disconnect(self, client, userdata, flags, result_code):
        print('Disconnected with result code {}.'.format(result_code))

    def send_audio_frames(self):
        stream = self.audio.open(format=pyaudio.paInt16, channels=CHANNELS,
                                 rate=FRAME_RATE, input=True,
                                 frames_per_buffer=CHUNK)

        print('Started broadcasting audio from device {}'
              ' on site {}'.format(self.audio_in, self.config.site))

        while True:
            frames = stream.read(CHUNK, exception_on_overflow=False)
            with io.BytesIO() as wav_buffer:
                with wave.open(wav_buffer, 'wb') as wav:
                    # pylint: disable=no-member
                    wav.setnchannels(CHANNELS)
                    wav.setsampwidth(SAMPLE_WIDTH)
                    wav.setframerate(FRAME_RATE)
                    wav.writeframes(frames)
                self.mqtt.publish(AUDIO_FRAME.format(self.config.site),
                                  wav_buffer.getvalue())



from pathlib import Path
from setuptools import setup, find_packages


SRC_ROOT = 'src'
BIN_ROOT = 'bin/'

# Read some information about the project from __about__.py
with (Path(SRC_ROOT) / 'hermes_audio_server' / '__about__.py').open('r') as fh:
    about = dict()
    exec(fh.read(), about)

with Path('README.md').open('r') as fh:
    long_description = fh.read()

with Path('requirements.txt').open('r') as fh:
    requirements = fh.read().splitlines()
    requirements = [requirement for requirement in requirements
                    if not requirement.startswith('#')]

binaries = [BIN_ROOT + about['__player__'], BIN_ROOT + about['__recorder__']]

setup(
    name=about['__project__'],
    version=about['__version__'],
    description=about['__description__'],
    long_description=long_description,
    long_description_content_type='text/markdown',
    license=about['__license__'],
    author=about['__author__'],
    author_email=about['__email__'],
    url=about['__github_url__'],
    project_urls={
        'Documentation': about['__doc_url__'],
        'Source': about['__github_url__'],
        'Tracker': about['__tracker_url__'],
    },
    packages=find_packages(SRC_ROOT),
    package_dir={'': SRC_ROOT},
    install_requires=requirements,
    include_package_data=True,
    zip_safe=False,
    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: Implementation :: CPython',
        'Topic :: Home Automation',
        'Topic :: Multimedia :: Sound/Audio :: Capture/Recording',
        'Topic :: Multimedia :: Sound/Audio :: Players'
    ],
    keywords=about['__keywords__'],
    scripts=binaries)



"""This module contains the main function run by the CLI commands
hermes-audio-player and hermes-audio-recorder.
"""
import sys
import traceback

from hermes_audio_server.__about__ import __version__
from hermes_audio_server.config import ServerConfig, DEFAULT_CONFIG
from hermes_audio_server.player import AudioPlayer
from hermes_audio_server.recorder import AudioRecorder


def main(command, verbose, version, config):
    """The main function run by the CLI command.

    Args:
        command (str): The command to run.
        verbose (bool): Use verbose output or not.
        version (bool): Print version information and exit.
        config (str): Configuration file.
    """
    try:
        print('{} {}'.format(command, __version__))
        if not version:
            if not config:
                config = DEFAULT_CONFIG

            if command == 'hermes-audio-player':
                server = AudioPlayer(ServerConfig.from_json_file(config),
                                     verbose)
            elif command == 'hermes-audio-recorder':
                server = AudioRecorder(ServerConfig.from_json_file(config),
                                       verbose)

            server.start()
    except FileNotFoundError as not_found:
        print('Configuration file {}.'.format(not_found))
    except KeyboardInterrupt:
        print('Shutting down {}...'.format(command))
        server.audio.terminate()
    except Exception:
        traceback.print_exc(file=sys.stdout)

    sys.exit(0)
